/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2020-2024. All rights reserved.
 */

#ifndef MOT_IMPL_TYPE_SLAMLOCALEGOODOM_H
#define MOT_IMPL_TYPE_SLAMLOCALEGOODOM_H

#include <cfloat>
#include <cmath>
#include <memory>
#include <iostream>
#include <vector>
#include <algorithm>
// #include "types.h"
#include "impl_type_float.h"
#include "impl_type_double.h"
// #include "impl_type_string.h"
#include "impl_type_uint8.h"
#include "impl_type_uint8_t.h"
#include "impl_type_uint16.h"
#include "impl_type_uint16_t.h"
#include "impl_type_uint32.h"
#include "impl_type_uint32_t.h"
#include "impl_type_uint64.h"
#include "impl_type_uint64_t.h"
#include "dpc/impl_type_slamlocalegoodom.h"
#include "perception/impl_type_lidarperception2fusion.h"
#include "impl_type_point3d.h"
#include "track.h"
#include "MOT_Wrapper.h"


// #include "lidar_processing.h"
struct Vector3f {
    ::Float x{ 0.0F };
    ::Float y{ 0.0F };
    ::Float z{ 0.0F };

    static bool IsPlane()
    {
        return true;
    }


    using IsEnumerableTag = void;
    template<typename F>
    void enumerate(F& fun)
    {
        fun(x);
        fun(y);
        fun(z);
    }

    template<typename F>
    void enumerate(F& fun) const
    {
        fun(x);
        fun(y);
        fun(z);
    }

    template<typename F>
    void enumerate_internal(F& fun)
    {
        fun("x", x);
        fun("y", y);
        fun("z", z);
    }

    template<typename F>
    void enumerate_internal(F& fun) const
    {
        fun("x", x);
        fun("y", y);
        fun("z", z);
    }

    friend bool operator==(const ::Vector3f& lhs, const ::Vector3f& rhs) noexcept
    {
        return (fabs(static_cast<double>(lhs.x - rhs.x)) < DBL_EPSILON) && (fabs(static_cast<double>(lhs.y - rhs.y)) < DBL_EPSILON) && (fabs(static_cast<double>(lhs.z - rhs.z)) < DBL_EPSILON);
    }
};
struct BBox2D {
    double xmin;
    double ymin;
    double xmax;
    double ymax;
};
struct BBox3D {
    double minx, miny, minz, maxx, maxy, maxz;
};
struct LidarPerceptionOutput {
    BBox2D bbox;
    Point3D position;
    BBox3D bbox3d;
    double theta;
    double length;
    double width;
    double height;
    std::vector<Point3D> polygon_point; // 3D points for polygon
    std::string type;
    std::string subtype;
    // a vector of Point3D representing a point cloud:
    std::vector<Point3D> point_cloud;
    double confidence;
    double time_stamp;
};

namespace mot {

    struct SLAMLocalEgoOdom {
        ::Double ego_odom_time_ns{ 0.0 };
        ::String status_inf;
        ::UInt8 origin_id{ 0U };
        ::Float yaw{ 0.0F };
        ::Float yaw_rate{ 0.0F };
        ::Vector3f position;
        ::Vector3f linear_velocity;

        friend bool operator==(const ::mot::SLAMLocalEgoOdom& lhs, const ::mot::SLAMLocalEgoOdom& rhs) noexcept
        {
            return (fabs(static_cast<double>(lhs.ego_odom_time_ns - rhs.ego_odom_time_ns)) < DBL_EPSILON) && (lhs.status_inf == rhs.status_inf) && (lhs.origin_id == rhs.origin_id) /*&& (fabs(static_cast<double>(lhs.yaw - rhs.yaw)) < DBL_EPSILON) && (lhs.position == rhs.position) && (lhs.linear_velocity == rhs.linear_velocity)*/;
        }
    };

    class SLAMDataBuffer {
    public:
        SLAMDataBuffer(size_t capacity);
        void addData(const mot::SLAMLocalEgoOdom& data);
        std::pair<size_t, size_t> findInterpolationPoints() const;

        // Function to interpolate between SLAMLocalEgoOdom data points
        void interpolate(
            const mot::SLAMLocalEgoOdom& before,
            const mot::SLAMLocalEgoOdom& after,
            mot::SLAMLocalEgoOdom* result) const;

        // doing the act
        void getInterpolatedData(double* time, mot::SLAMLocalEgoOdom* interpolatedData);
        std::pair<size_t, size_t> findInterpolationPoints();

    private:
        std::vector<mot::SLAMLocalEgoOdom> buffer;
        size_t head;
        size_t size;
        double targetTime; // Time you want to interpolate at
    };

    class InputWrapper {
        public:
            InputWrapper(size_t bufferSize);
            // void MotWrapper(std::vector<LidarPerceptionOutput>* Det_Objs, dpc::SLAMLocalEgoOdom &vehicle_odom_vec_Temp, double time);
            void MotWrapper(Eigen::MatrixXf &bboxes_3d, std::vector<int> &best_object, dpc::SLAMLocalEgoOdom &vehicle_odom_vec_Temp, double time);
        private:
            void _wrapper(dpc::SLAMLocalEgoOdom* vehicle_odom_vec_Temp, mot::SLAMLocalEgoOdom* vehicle_odom_vec);
            uint32_t Kinematics_Relative2OverGround( mot::SLAMLocalEgoOdom* vehicle_odom_vec,
                LidarPerceptionOutput* pCurrTrkbl, const float f32deltaX);  
            SLAMDataBuffer buffer;
            double timestamp;
            uint32_t Usuccess;
            const float f32deltaX;
            mot::SLAMLocalEgoOdom vehicle_odom_vec;
            perception::LiDARPerception2Fusion MOT_output_objs;
            dpc::SLAMLocalEgoOdom vehicle_odom_vec_Temp;
            std::vector<LidarPerceptionOutput> RuleBasedResult; 
    };
} // namespace mot

// double Main_MOT_Wrapper(std::vector<LidarPerceptionOutput>* Det_Objs, dpc::SLAMLocalEgoOdom &vehicle_odom_vec_Temp);
void output_wrapper(std::vector<perception::LiDARPerception2Fusion>& MOT_output_objs, std::vector<lidar_tracker::TrackPtr>& Track_List);
void Kinematics_local2overground(std::vector<perception::LiDARPerception2Fusion>& MOT_output_objs, dpc::SLAMLocalEgoOdom* vehicle_odom_vec_Temp);



#endif // MOT_IMPL_TYPE_SLAMLOCALEGOODOM_H
